"""
.. module: historical.common.sqs
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import logging
import uuid
import random

import boto3

from historical.constants import CURRENT_REGION

logging.basicConfig()
log = logging.getLogger('historical')
log.setLevel(logging.INFO)


def chunks(event_list, chunk_size):
    """Yield successive n-sized chunks from the event list."""
    for i in range(0, len(event_list), chunk_size):
        yield event_list[i:i + chunk_size]


def get_queue_url(queue_name):
    """Get the URL of the SQS queue to send events to."""
    client = boto3.client("sqs", CURRENT_REGION)
    queue = client.get_queue_url(QueueName=queue_name)

    return queue["QueueUrl"]


def make_sqs_record(event, delay_seconds=0):
    """Get a dict with the components required for SQS"""
    return {
        "Id": uuid.uuid4().hex,
        "DelaySeconds": delay_seconds,
        "MessageBody": event
    }


def get_random_delay(max):
    """Gets a randomized number between 0 and the max number in seconds for
       how long a message in SQS should be delayed.

       900 seconds (15 min) is the maximum permitted by SQS.
    :return:
    """
    return random.randint(0, max)


def produce_events(events, queue_url, batch_size=10, randomize_delay=0):
    """
    Efficiently sends events to the SQS event queue.

    Note: SQS has a max size of 10 items.  Please be aware that this can make the messages go past size -- even
    with shrinking messages!

    Events can get randomized delays, maximum of 900 seconds. Set that in `randomize_delay`
    :param events:
    :param queue_url:
    :param batch_size:
    :param randomize_delay:
    """
    client = boto3.client('sqs', region_name=CURRENT_REGION)

    for chunk in chunks(events, batch_size):
        records = [make_sqs_record(event, delay_seconds=get_random_delay(randomize_delay)) for event in chunk]

        client.send_message_batch(Entries=records, QueueUrl=queue_url)


def group_records_by_type(records, update_events):
    """Break records into two lists; create/update events and delete events."""
    update_records, delete_records = [], []
    for r in records:
        if r.get("detail-type", "") == "Scheduled Event":
            log.error("[X] Received a Scheduled Event in the Queue... Please check that your environment is set up"
                      " correctly.")
            continue

        # Ignore SQS junk messages (like subscription notices and things):
        if not r.get("detail"):
            continue

        # Do not capture error events:
        if not r["detail"].get("errorCode"):
            if r['detail']['eventName'] in update_events:
                update_records.append(r)
            else:
                delete_records.append(r)

    return update_records, delete_records



"""
.. module: historical.s3.collector
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import logging
from itertools import groupby

from botocore.exceptions import ClientError
from pynamodb.exceptions import PynamoDBConnectionError
from raven_python_lambda import RavenLambdaWrapper
from cloudaux.orchestration.aws.s3 import get_bucket

from historical.common.sqs import group_records_by_type
from historical.constants import HISTORICAL_ROLE, CURRENT_REGION, LOGGING_LEVEL
from historical.common import cloudwatch
from historical.common.util import deserialize_records
from historical.s3.models import CurrentS3Model, VERSION

logging.basicConfig()
log = logging.getLogger('historical')
log.setLevel(LOGGING_LEVEL)


UPDATE_EVENTS = [
    'DescribeBucket',   # Polling event
    'DeleteBucketCors',
    'DeleteBucketLifecycle',
    'DeleteBucketPolicy',
    'DeleteBucketReplication',
    'DeleteBucketTagging',
    'DeleteBucketWebsite',
    'CreateBucket',
    'PutBucketAcl',
    'PutBucketCors',
    'PutBucketLifecycle',
    'PutBucketPolicy',
    'PutBucketLogging',
    'PutBucketNotification',
    'PutBucketReplication',
    'PutBucketTagging',
    'PutBucketRequestPayment',
    'PutBucketVersioning',
    'PutBucketWebsite'
]


DELETE_EVENTS = [
    'DeleteBucket',
]


def create_delete_model(record):
    """Create an S3 model from a record."""
    arn = 'arn:aws:s3:::{}'.format(cloudwatch.filter_request_parameters('bucketName', record))
    log.debug('[-] Deleting Dynamodb Records. Hash Key: {arn}'.format(arn=arn))

    data = {
        'arn': arn,
        'principalId': cloudwatch.get_principal(record),
        'userIdentity': cloudwatch.get_user_identity(record),
        'accountId': record['account'],
        'eventTime': record['detail']['eventTime'],
        'BucketName': cloudwatch.filter_request_parameters('bucketName', record),
        'Region': cloudwatch.get_region(record),
        'Tags': {},
        'configuration': {},
        'eventSource': record['detail']['eventSource'],
        'version': VERSION
    }

    return CurrentS3Model(**data)


def process_delete_records(delete_records):
    """Process the requests for S3 bucket deletions"""
    for r in delete_records:
        arn = 'arn:aws:s3:::{}'.format(r['detail']['requestParameters']['bucketName'])

        # Need to check if the event is NEWER than the previous event in case
        # events are out of order. This could *possibly* happen if something
        # was deleted, and then quickly re-created. It could be *possible* for the
        # deletion event to arrive after the creation event. Thus, this will check
        # if the current event timestamp is newer and will only delete if the deletion
        # event is newer.
        try:
            log.debug('[-] Deleting bucket: {}'.format(arn))
            model = create_delete_model(r)
            model.save(condition=(CurrentS3Model.eventTime <= r['detail']['eventTime']))
            model.delete()

        except PynamoDBConnectionError as pdce:
            log.warn("[?] Unable to delete bucket: {}. Either it doesn't exist, or this deletion event is stale "
                     "(arrived before a NEWER creation/update). The specific exception is: {}".format(arn, pdce))


def process_update_records(update_records):
    """Process the requests for S3 bucket update requests"""
    events = sorted(update_records, key=lambda x: x['account'])

    # Group records by account for more efficient processing
    for account_id, events in groupby(events, lambda x: x['account']):
        events = list(events)

        # Grab the bucket names (de-dupe events):
        buckets = {}
        for e in events:
            # If the creation date is present, then use it:
            bucket_event = buckets.get(e['detail']['requestParameters']['bucketName'], {
                'creationDate': e['detail']['requestParameters'].get('creationDate')
            })
            bucket_event.update(e['detail']['requestParameters'])

            buckets[e['detail']['requestParameters']['bucketName']] = bucket_event
            buckets[e['detail']['requestParameters']['bucketName']]['eventDetails'] = e

        # Query AWS for current configuration
        for b, item in buckets.items():
            log.debug('[~] Processing Create/Update for: {}'.format(b))
            # If the bucket does not exist, then simply drop the request --
            # If this happens, there is likely a Delete event that has occurred and will be processed soon.
            try:
                bucket_details = get_bucket(b,
                                            account_number=account_id,
                                            include_created=(item.get('creationDate') is None),
                                            assume_role=HISTORICAL_ROLE,
                                            region=CURRENT_REGION)
                if bucket_details.get('Error'):
                    log.error('[X] Unable to fetch details about bucket: {}. '
                              'The error details are: {}'.format(b, bucket_details['Error']))
                    continue

            except ClientError as ce:
                if ce.response['Error']['Code'] == 'NoSuchBucket':
                    log.warning('[?] Received update request for bucket: {} that does not '
                                'currently exist. Skipping.'.format(b))
                    continue

                # Catch Access Denied exceptions as well:
                if ce.response['Error']['Code'] == 'AccessDenied':
                    log.error('[X] Unable to fetch details for S3 Bucket: {} in {}. Access is Denied. '
                              'Skipping...'.format(b, account_id))
                    continue
                raise Exception(ce)

            # Pull out the fields we want:
            data = {
                'arn': 'arn:aws:s3:::{}'.format(b),
                'principalId': cloudwatch.get_principal(item['eventDetails']),
                'userIdentity': cloudwatch.get_user_identity(item['eventDetails']),
                'accountId': account_id,
                'eventTime': item['eventDetails']['detail']['eventTime'],
                'BucketName': b,
                'Region': bucket_details.pop('Region'),
                # Duplicated in top level and configuration for secondary index
                'Tags': bucket_details.pop('Tags', {}) or {},
                'eventSource': item['eventDetails']['detail']['eventSource'],
                'version': VERSION
            }

            # Remove the fields we don't care about:
            del bucket_details['Arn']
            del bucket_details['GrantReferences']
            del bucket_details['_version']
            del bucket_details['Name']

            if not bucket_details.get('CreationDate'):
                bucket_details['CreationDate'] = item['creationDate']

            data['configuration'] = bucket_details

            current_revision = CurrentS3Model(**data)
            current_revision.save()


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical S3 event collector.

    This collector is responsible for processing CloudWatch events and polling events.
    """
    records = deserialize_records(event['Records'])

    # Split records into two groups, update and delete.
    # We don't want to query for deleted records.
    update_records, delete_records = group_records_by_type(records, UPDATE_EVENTS)

    log.debug('[@] Processing update records...')
    process_update_records(update_records)
    log.debug('[@] Completed processing of update records.')

    log.debug('[@] Processing delete records...')
    process_delete_records(delete_records)
    log.debug('[@] Completed processing of delete records.')

    log.debug('[@] Successfully updated current Historical table')



"""
.. module: historical.common.util
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json


def deserialize_records(records):
    """
    This properly deserializes records depending on where they came from:
        - SQS
        - SNS
    """
    native_records = []
    for r in records:
        parsed = json.loads(r['body'])

        # Is this a DynamoDB stream event?
        if isinstance(parsed, str):
            native_records.append(json.loads(parsed))

        # Is this a subscription message from SNS? If so, skip it:
        elif parsed.get('Message') and parsed.get('UnsubscribeURL'):
            continue

        # Is this from SNS (cross-region request -- SNS messages wrapped in SQS message) -- or an SNS proxied message?
        elif parsed.get('Message'):
            native_records.append(json.loads(parsed['Message']))

        else:
            native_records.append(parsed)

    return native_records


def pull_tag_dict(data):
    """This will pull out a list of Tag Name-Value objects, and return it as a dictionary.

    :param data: The dict collected from the collector.
    :returns dict: A dict of the tag names and their corresponding values.
    """
    # If there are tags, set them to a normal dict, vs. a list of dicts:
    tags = data.pop('Tags', {}) or {}
    if tags:
        proper_tags = {}
        for t in tags:
            proper_tags[t['Key']] = t['Value']

        tags = proper_tags

    return tags



"""
.. module: historical.constants
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import logging
import os

log_levels = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG
}


def extract_log_level_from_environment(k, default):
    return log_levels.get(os.environ.get(k)) or int(os.environ.get(k, default))


# 24 hours in seconds is the default
TTL_EXPIRY = int(os.environ.get('TTL_EXPIRY', 86400))

# By default, don't randomize the pollers (tasker or collector -- same env var):
RANDOMIZE_POLLER = int(os.environ.get('RANDOMIZE_POLLER', 0))

CURRENT_REGION = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')
HISTORICAL_ROLE = os.environ.get('HISTORICAL_ROLE', 'Historical')
POLL_REGIONS = os.environ.get('POLL_REGIONS', 'us-east-1').split(",")
PROXY_REGIONS = os.environ.get('PROXY_REGIONS', 'us-east-1').split(",")
REGION_ATTR = os.environ.get('REGION_ATTR', 'Region')
LOGGING_LEVEL = extract_log_level_from_environment('LOGGING_LEVEL', logging.INFO)
EVENT_TOO_BIG_FLAG = "event_too_big"



"""
.. module: historical.attributes
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json
import decimal

from pynamodb.attributes import (
    Attribute,
    MapAttribute,
    ListAttribute,
    BooleanAttribute,
    NumberAttribute,
)

import pynamodb
from pynamodb.constants import STRING, NUMBER

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


class HistoricalUnicodeAttribute(Attribute):
    """
    A Historical unicode attribute.
    Replaces '' with '<empty>' during serialization and correctly deserialize '<empty>' to ''
    """
    attr_type = STRING

    def serialize(self, value):
        """
        Returns a unicode string
        """
        if value is None or not len(value):
            return '<empty>'
        return value

    def deserialize(self, value):
        if value == '<empty>':
            return ''
        return value


class EventTimeAttribute(Attribute):
    """
    An attribute for storing a UTC Datetime or iso8601 string
    """
    attr_type = STRING

    def serialize(self, value):
        """
        Takes a datetime object and returns a string
        """
        if isinstance(value, str):
            return value
        return value.strftime(DATETIME_FORMAT)


def decimal_default(obj):
    if isinstance(obj, decimal.Decimal):
        if obj % 1:
            return float(obj)
        return int(obj)
    raise TypeError


class HistoricalDecimalAttribute(Attribute):
    """
    A number attribute
    """
    attr_type = NUMBER

    def serialize(self, value):
        """
        Encode numbers as JSON
        """
        return json.dumps(value, default=decimal_default)

    def deserialize(self, value):
        """
        Decode numbers from JSON
        """
        return json.loads(value)


pynamodb.attributes.SERIALIZE_CLASS_MAP = {
    dict: MapAttribute(),
    list: ListAttribute(),
    set: ListAttribute(),
    bool: BooleanAttribute(),
    float: NumberAttribute(),
    int: NumberAttribute(),
    str: HistoricalUnicodeAttribute(),
    decimal.Decimal: HistoricalDecimalAttribute()
}



"""
.. module: historical.common.accounts
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
.. author:: Mike Grima <mgrima@netflix.com>
"""
import os

from swag_client.backend import SWAGManager
from swag_client.util import parse_swag_config_options


def parse_boolean(value):
    """Simple function to get a boolean value from string."""
    if not value:
        return False

    if str(value).lower() == 'true':
        return True

    return False


def get_historical_accounts():
    """Fetches valid accounts from SWAG if enabled or a list accounts."""
    if os.environ.get('SWAG_BUCKET', False):
        swag_opts = {
            'swag.type': 's3',
            'swag.bucket_name': os.environ['SWAG_BUCKET'],
            'swag.data_file': os.environ.get('SWAG_DATA_FILE', 'accounts.json'),
            'swag.region': os.environ.get('SWAG_REGION', 'us-east-1')
        }
        swag = SWAGManager(**parse_swag_config_options(swag_opts))
        search_filter = "[?provider=='aws'] && [?owner=='{}']".format(os.environ['SWAG_OWNER'])

        if parse_boolean(os.environ.get('TEST_ACCOUNTS_ONLY')):
            search_filter += " && [?environment=='test']"

        accounts = swag.get_service_enabled('historical', search_filter=search_filter)
    else:
        accounts = os.environ['ENABLED_ACCOUNTS']

    return accounts



"""
Historical
==========

Allows for the tracking of AWS configuration data across accounts/regions/technologies.

"""
import os.path
import sys

from setuptools import setup, find_packages

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

# When executing the setup.py, we need to be able to import ourselves, this
# means that we need to add the src/ directory to the sys.path.
sys.path.insert(0, ROOT)

about = {}
with open(os.path.join(ROOT, "historical", "__about__.py")) as f:
    exec(f.read(), about)


install_requires = [
    'boto3==1.7.79',   # moto is currently broken with the latest and greatest version :(
    'cloudaux>=1.4.14',
    'click>=6.7',
    'pynamodb>=3.1.0',
    'deepdiff>=3.3.0',
    'raven-python-lambda>=0.1.7',
    'marshmallow>=2.13.5',
    'swag-client>=0.3.0',
    'python-dateutil==2.6.1'
]

tests_require = [
    'pytest==3.1.3',
    'moto>=1.3.2',
    'coveralls==1.1',
    'factory-boy==2.9.2'
]


setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    description=about["__summary__"],
    long_description='See README.md',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require
    },
    entry_points={
        'console_scripts': [
            'historical = historical.cli:cli',
        ]
    },
    keywords=['aws', 'account_management'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
